/*
 * cfgtool - A Configurator tool for ROSAND Technologies products.
 *
 * Inspired from AVRDUDE program.
 *
 * Copyright (C) 2000-2005  Brian S. Dean <bsd@bsdhome.com>
 * Copyright 2007-2009 Joerg Wunsch <j@uriah.heep.sax.de>
 * Copyright (c) 2011, Rosand Technologies
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/time.h>
#include <unistd.h>

#include "config.h"
#include "pgm.h"
#include "swra327.h"
#include "serial.h"

struct pdata
{
  unsigned char *flash_pagecache;
  unsigned int flash_pagesize;
};

#define PDATA(pgm) ((struct pdata *)(pgm->cookie))

static void swra327_setup(PROGRAMMER * pgm)
{
  if ((pgm->cookie = malloc(sizeof(struct pdata))) == 0) {
    fprintf(stderr,
	    "%s: swra327_setup(): Out of memory allocating private data\n",
	    progname);
    return;
  }
  memset(pgm->cookie, 0, sizeof(struct pdata));
}

static void swra327_teardown(PROGRAMMER * pgm)
{
  free(pgm->cookie);
}

static int swra327_send(PROGRAMMER * pgm, char * data, size_t len)
{
  unsigned char buf[256 + 4];              /* max MESSAGE_BODY of 255 bytes, 4 bytes overhead */
  unsigned char crc = 0;
  int i;

  memset(buf, 0xff, sizeof(buf));
  buf[0] = START_OF_FRAME;
  buf[1] = len-1;
  buf[2] = TOKEN;
  memcpy(buf+3, data, len);

  // calculate the XOR checksum
  for (i=0; i<3+len; i++)
	  crc ^= buf[i];
  buf[i++] = crc;

  return serial_send(&pgm->fd, buf, i);
}

static int swra327_recv(PROGRAMMER * pgm, char * buf, size_t len)
{
  return serial_recv(&pgm->fd, (unsigned char *)buf, len);
}

static int swra327_drain(PROGRAMMER * pgm, int display)
{
  return serial_drain(&pgm->fd, display);
}

static int swra327_vfy_cmd_sent(PROGRAMMER * pgm, unsigned char *data,  char * errmsg)
{
  unsigned char buf[275];
  unsigned char *tmp = buf;
  unsigned char c, crc = 0;
  int i;
  unsigned char len;

  if (swra327_recv(pgm, tmp, 2) < 0) {
    fprintf(stderr, "%s: %s: programmer did not respond to command.\n",
            progname, errmsg);
    return -1;
  }
  len = *(tmp+1);
  tmp +=2;

  swra327_recv(pgm, tmp, len+3);

  for (i=0, tmp=buf; i < len+4; i++) {
	  crc ^= *tmp++;
  }
  if (crc != *tmp) {
	  if (verbose)
		  fprintf(stderr, "%s: %s: CRC response incorrect.\n", progname, errmsg);
	  return -1;
  }

  tmp = buf + 4;

  if ((c = *(tmp++)) != 0) {
	  if (verbose)
		  fprintf(stderr, "%s: %s: ", progname, errmsg);
	  switch(c) {
	  case 0x01:
		  if (verbose)
			  fprintf(stderr, "command not succeed.\n");
		  break;
	  case 0x02:
		  if (verbose)
			  fprintf(stderr, "CRC incorrect.\n");
		  break;
	  default:
		  if (verbose)
			  fprintf(stderr, "(code = 0x%02x)\n.",c);
	  }
	  return -2;
  }

  len--;
  for (i=0; i<len; i++) {
	  *data++ = *tmp++;
  }
  return len;
}

static int swra327_leds(PROGRAMMER *pgm, unsigned char command, unsigned char leds, unsigned char *data)
{
  unsigned char buffer[10];
  unsigned char *tmp = buffer;
  int result;

  *tmp++ = CMD_LEDS;
  *tmp++ = command;
  *tmp++ = leds;

  swra327_send(pgm, buffer, tmp - buffer);

  result = swra327_vfy_cmd_sent(pgm, data, "leds command");
  if (result < 0) {
	  fprintf(stderr, "%s: error: leds command\n", progname);
	  return -1;
  }
  return 0;
}

static int swra327_err_led(PROGRAMMER *pgm, int command)
{
  return swra327_leds(pgm, command, LEDS_ERR, NULL);
}

/*
 * issue the 'chip erase' command to the device
 */
static int swra327_chip_erase(PROGRAMMER * pgm, PART * p)
{
  unsigned char c = CMD_ERASE;
  int result;

  swra327_send(pgm, &c, 1);
  result = swra327_vfy_cmd_sent(pgm, NULL, "chip erase");
  if (result < 0) {
	  fprintf(stderr, "%s: error: chip erase\n", progname);
	  return -1;
  }

  /*
   * swra327 firmware may not delay long enough
   */
  usleep (p->chip_erase_delay);

  return 0;
}

static void swra327_enter_prog_mode(PROGRAMMER * pgm)
{
  unsigned char c = CMD_HAND_SHAKE;
  int result;
  unsigned char buf[256];

  swra327_send(pgm, &c, 1);

  memset(buf, 0, sizeof(buf));
  result = swra327_vfy_cmd_sent(pgm, buf, "enter prog mode");
  if (result < 0) {
	  fprintf(stderr, "%s: error: enter prog mode\n", progname);
	  return;
  }
}

/*
 * issue the 'program enable' command to the device
 */
static int swra327_program_enable(PROGRAMMER * pgm, PART * p)
{
  int result;
  unsigned char c = CMD_ENABLE;

  swra327_send(pgm, &c, 1);

  result = swra327_vfy_cmd_sent(pgm, NULL, "prog enable");
  if (result < 0) {
	  fprintf(stderr, "%s: error: prog enable\n", progname);
	  return -1;
  }
  return 0;
}

static void swra327_leave_prog_mode(PROGRAMMER * pgm)
{
  unsigned char c = CMD_END;

  swra327_err_led(pgm, CMD_LEDS_ON);

  swra327_send(pgm, &c, 1);

  swra327_vfy_cmd_sent(pgm, NULL, "leave prog mode");
}


static int swra327_initialize(PROGRAMMER * pgm, PART * p)
{
  unsigned char c = SB_FORCE_BOOT;

  serial_send(&pgm->fd, &c, 1);

  swra327_enter_prog_mode(pgm);

  return 0;
}

static void swra327_disable(PROGRAMMER * pgm)
{
  /* Do nothing. */

  return;
}

static void swra327_enable(PROGRAMMER * pgm)
{
  /* Do nothing. */

  return;
}

/*
 * transmit a device command and return the results; 'cmd' and
 * 'res' must point to at least a 4 byte data buffer
 */
static int swra327_cmd(PROGRAMMER * pgm, unsigned char *cmd,
		unsigned char *res)
{
  int result;

  swra327_send(pgm, &cmd[1], cmd[0]);

  result = swra327_vfy_cmd_sent(pgm, NULL, "swra327_cmd");
  if (result < 0) {
	  return -1;
  }

  return result;
}

static int swra327_open(PROGRAMMER * pgm, char * port)
{
  /*
   *  If baudrate was not specified use 19.200 Baud
   */
  if(pgm->baudrate == 0) {
    pgm->baudrate = 19200;
  }

  strcpy(pgm->port, port);
  serial_open(port, pgm->baudrate, &pgm->fd);

  /*
   * drain any extraneous input
   */
  swra327_drain (pgm, 0);

  return 0;
}

static void swra327_close(PROGRAMMER * pgm)
{
  swra327_leave_prog_mode(pgm);

  serial_close(&pgm->fd);
  pgm->fd.ifd = -1;
}

static void swra327_display(PROGRAMMER * pgm, const char * p)
{
  return;
}

static int swra327_load(PROGRAMMER * pgm, PART * p, MEM *mem, int index, long n_bytes)
{
  int cmdlen, result;
  unsigned char *tmp, buf[275];
  unsigned char *cache_ptr = NULL;

  if (verbose >= 2)
	  fprintf(stderr, "%s: swra327_load(..., %s, 0x%lx, ...)\n",
			  progname, mem->desc, n_bytes);

  tmp = buf;
  *tmp++ = CMD_READ;
  *tmp++ = mem->pollindex;
  if(index) *tmp++ = index;
  *tmp++ = n_bytes;
  cmdlen = tmp-buf;

  result = swra327_send(pgm, buf, cmdlen);

  if (result < 0) {
	  fprintf(stderr,
			  "%s: swra327_paged_load(): timeout/error communicating with product\n",
			  progname);
	  return -1;
  }

  // attempt to get the data
  result = swra327_vfy_cmd_sent(pgm, buf, "verify read command");

  // check read back
  if (result < 0) {
	  fprintf(stderr, "%s: swra327_load(): error reading data\n",
			  progname);
	  return -1;
  }
  tmp = buf;
  cache_ptr = mem->buf;
  do {
	  *cache_ptr++ = *tmp++;
  } while (tmp-buf < n_bytes);

  return n_bytes;
}

static int swra327_write(PROGRAMMER * pgm, PART * p, MEM *mem, int index, long n_bytes)
{
  int cmdlen, result;
  unsigned char c, buf[275];
  unsigned char *tmp;

  if (verbose >= 2)
	  fprintf(stderr, "%s: swra327_write(.., %s, 0x%lx, ...)\n",
			  progname, mem->desc, n_bytes);
  tmp = buf;
  *tmp++ = CMD_WRITE;
  *tmp++ = mem->pollindex;
  if(index)  {
	  *tmp++ = index;
	  *tmp++ = mem->size;
  }
  cmdlen = n_bytes + (tmp-buf);
  memcpy(tmp, mem->buf, n_bytes);
  result = swra327_send(pgm, buf, cmdlen);
  if (result < 0) {
	  fprintf(stderr,
			  "%s: swra327_paged_write(): "
			  "timeout/error communicating with programmer\n",
			  progname);
	  return -1;
  }

  result = swra327_vfy_cmd_sent(pgm, &c, "swra327_paged_write");
  return n_bytes;
}

void swra327_initpgm(PROGRAMMER * pgm)
{
  strcpy(pgm->type, "serial");

  /*
   * mandatory functions
   */
  pgm->initialize     = swra327_initialize;
  pgm->display        = swra327_display;
  pgm->enable         = swra327_enable;
  pgm->disable        = swra327_disable;
  pgm->program_enable = swra327_program_enable;
  pgm->chip_erase     = swra327_chip_erase;
  pgm->cmd            = swra327_cmd;
  pgm->open           = swra327_open;
  pgm->close          = swra327_close;

  /*
   * optional functions
   */
  pgm->paged_write    = swra327_write;
  pgm->paged_load     = swra327_load;
  pgm->setup          = swra327_setup;
  pgm->teardown       = swra327_teardown;
  pgm->err_led        = swra327_err_led;
}
